@@ -18,59 +18,160 @@ from . import phrasing
 # Strict equality #
 #-----------------#
 
-def strict_equality_checker(val, ref):
+def report_unequal(val, ref):
+    """
+    Returns a status/explanation dictionary reporting that two values
+    are unequal, including a report on their differences, which has
+    different forms depending on the values themselves.
+    """
+    if isinstance(val, str) and isinstance(ref, str):
+        val_lines = val.splitlines(keepends=True)
+        ref_lines = ref.splitlines(keepends=True)
+
+        diff = html_tools.html_diff_table(
+            val_lines,
+            ref_lines,
+            'Actual value',
+            'Corresponding lines (if any) in expected value'
+        )
+    else:
+        rval = repr(val)
+        rexp = repr(ref)
+        if max(len(rval), len(rexp)) < 50:
+            diff = "Actual: {}<br>\nExpected: {}".format(
+                repr(val),
+                repr(ref)
+            )
+        else:
+            diff = "Actual:<br>\n{}<br><br>\nExpected:<br>\n{}".format(
+                html_tools.dynamic_html_repr(val),
+                html_tools.dynamic_html_repr(ref)
+            )
+
+    return {
+        "status": "failed",
+        "explanation": (
+            "Actual value is different from the reference"
+            " value:<br>\n{}"
+        ).format(diff)
+    }
+
+
+def report_equal(val, ref):
+    """
+    Returns a status/explanation dictionary reporting that two values
+    are equivalent.
+    """
+    val_repr = repr(val)
+    if isinstance(val, str) and '\n' in val:
+        val_repr = "'''\\\n{}'''".format(val)
+    return {
+        "status": "accomplished",
+        "explanation": (
+            "Actual value is the same as the reference value:<br>\n{}"
+        ).format(
+            html_tools.html_output_details(
+                val_repr, # TODO: Pretty printing here
+                "Value tested"
+            )
+        )
+    }
+
+
+def strict_equality_checker(val, ref, memo=None):
     """
     A checker to be used with OutputTest or ValueTest goals that ensures
     the output or value being checked is strictly equal to the reference
     output or value. If there is a difference, some kind of diff is
     printed, depending on the types of the values.
+
+    Handles recursive structures composed of tuples, lists, sets, and/or
+    dictionaries, but may choke on recursive structures stored in custom
+    objects, in which case those objects will only be considered equal
+    if they are the same object.
     """
-    if type(val) == type(ref) and val == ref:
-        val_repr = repr(val)
-        if isinstance(val, str) and '\n' in val:
-            val_repr = val
+    if memo is None:
+        memo = set()
+
+    mkey = (id(val), id(ref))
+    # If we're already in the process of comparing these two objects to
+    # each other, and we come across another comparison of the two, we
+    # can safely return True for the sub-comparison, and let other
+    # comparisons determine if they're equal or not.
+    if mkey in memo:
         return {
             "status": "accomplished",
-            "explanation": (
-                "Actual value is the same as the reference value.<br>\n{}"
-            ).format(
-                html_tools.html_output_details(
-                    val_repr, # TODO: Pretty printing here
-                    "Value tested"
-                )
-            )
+            "explanation": "Recursive comparison."
         }
-    else:
-        if isinstance(val, str) and isinstance(ref, str):
-            val_lines = val.splitlines(keepends=True)
-            ref_lines = ref.splitlines(keepends=True)
 
-            diff = html_tools.html_diff_table(
-                val_lines,
-                ref_lines,
-                'Actual output',
-                'Corresponding lines (if any) in expected output'
-            )
+    # Past this point, recursion might happen
+    memo.add(mkey)
+
+    if type(val) != type(ref):
+        return report_unequal(val, ref)
+    elif isinstance(val, (int, float, complex, type(None), bool, str)):
+        if val == ref:
+            return report_equal(val, ref)
         else:
-            rval = repr(val)
-            rexp = repr(ref)
-            if max(len(rval), len(rexp)) < 50:
-                diff = "Actual: {}<br>\nExpected: {}".format(
-                    repr(val),
-                    repr(ref)
-                )
+            return report_unequal(val, ref)
+    elif isinstance(val, (tuple, list)):
+        if len(val) != len(ref):
+            return report_unequal(val, ref)
+        else:
+            test = all(
+                strict_equality_checker(v, r, memo)\
+                    ["status"] == "accomplished" # noqa E211
+                for (v, r) in zip(val, ref)
+            )
+            if test:
+                return report_equal(val, ref)
             else:
-                diff = "Actual:<br>\n{}<br><br>\nExpected:<br>\n{}".format(
-                    html_tools.dynamic_html_repr(val),
-                    html_tools.dynamic_html_repr(ref)
-                )
+                return report_unequal(val, ref)
+    elif isinstance(val, (set)):
+        sv = sorted(val)
+        sr = sorted(ref)
+        test = strict_equality_checker(sv, sr, memo)
+        if test["status"] == "accomplished":
+            return report_equal(val, ref)
+        else:
+            return report_unequal(val, ref)
+    elif isinstance(val, (dict)):
+        test = strict_equality_checker(
+            sorted(val.keys()),
+            sorted(ref.keys()),
+            memo
+        )
+        if test["status"] != "accomplished":
+            return report_unequal(val, ref)
 
-        return {
-            "status": "failed",
-            "explanation": (
-                "Actual value is different from the reference value:<br>\n{}"
-            ).format(diff)
-        }
+        failed = False
+        for key in val: # keys must be same at this point
+            test = strict_equality_checker(val[key], ref[key], memo)
+            if test["status"] != "accomplished":
+                failed = True
+                break
+
+        if failed:
+            return report_unequal(val, ref)
+        else:
+            return report_equal(val, ref)
+    else:
+        # Some kind of object that we don't know about; try calling
+        # __eq__ if it has it, an upon a recursion error, simply
+        # compare IDs (which is Python default behavior when __eq__
+        # isn't present).
+        if hasattr(val, "__eq__"):
+            try:
+                test = val == ref
+            except RecursionError:
+                test = val is ref
+        else:
+            test = val is ref
+
+        if test:
+            return report_equal(val, ref)
+        else:
+            return report_unequal(val, ref)
 
 
 #---------------------------#
@@ -1368,16 +1469,16 @@ def sequence_diff(
 
         # Slower than sets but preserves ordering in each list
         matching = [
-            val[idx]
+            list(val)[idx]
             for idx in correspondence
         ]
         missing = [
-            ref[idx]
+            list(ref)[idx]
             for idx in range(len(ref))
             if idx not in matched_indices
         ]
         extra = [
-            val[idx]
+            list(val)[idx]
             for idx in range(len(val))
             if idx not in correspondence
         ]
@@ -1763,6 +1864,174 @@ def make_structure_comparator(
     return compare_structures
 
 
+#------------------#
+# Image comparison #
+#------------------#
+
+def diff_anim_frames(val, ref, steps=10):
+    """
+    Creates and returns a list of PIL.Image objects representing
+    animation frames that fade between the two given images. A custom
+    number of fade steps may be provided, which specifies the number of
+    intermediate frames, so the result will have that many frames plus
+    two. The fade is linear.
+    """
+    import PIL.Image # we import here and let a potential error bubble out
+    # TODO: Draw label text on anim frames!
+
+    result = [val]
+    for step in range(1, steps + 1):
+        t = step / (steps + 1)
+        result.append(PIL.Image.blend(val, ref, t))
+    result.append(ref)
+    return result
+
+
+def make_image_comparator(
+    allowed_differences=0.03,
+    partial_allowed=0.5,
+    similarity_threshold=15
+):
+    """
+    Returns a comparison function suitable for use with 'image' context
+    slots. It checks pixels in the two images (which it checks are
+    actually images and are the same size) and returns a status of
+    accomplished if the fraction of the area of the image in which pixels
+    differ is less than or equal to the given allowed fraction. Pixels
+    which have a 255-basis-RGB-space Euclidean distance (TODO: better
+    distance metric) below the given similarity threshold are counted as
+    1/2 of a pixel difference instead of a full pixel difference. (TODO:
+    Better controls for this?). A partially-complete result is returned
+    as long as the fractional area of different pixels is no more than
+    the given partial_allowed value.
+    """
+    import PIL.Image # we import here and let a potential error bubble out
+
+    def compare_images(val, ref):
+        """
+        Compares two PILlow images, ensuring that the total percentage of
+        pixels which differ between the two images is below a threshold.
+        Pixels which differ only slightly will count as 1/2 a pixel in
+        terms of the area difference computation. Images of different
+        sizes will always be treated as totally different (TODO: More
+        nuance there?).
+        """
+        if not isinstance(ref, PIL.Image.Image):
+            raise TypeError(
+                f"Reference image capture failed (got a {type(ref)},"
+                f" not a PIL.Image)"
+            )
+        if not isinstance(val, PIL.Image.Image):
+            return {
+                "status": "failed",
+                "explanation": (
+                    f"Image capture failed (got a {type(val)}, not an"
+                    f" image)."
+                )
+            }
+
+        if val.size != ref.size:
+            return {
+                "status": "failed",
+                "explanation": (
+                    f"Your image ({val.width}x{val.height}) did not"
+                    f" have the same size as the solution image"
+                    f" ({ref.width}x{ref.height})."
+                )
+            }
+
+        # Should be 0-255 RGB data
+        # TODO: Verify mode
+        val_pixels = val.getdata()
+        ref_pixels = ref.getdata()
+        diff_area = 0
+        half_diff_area = 0
+        for index in range(len(val_pixels)):
+            px = val_pixels[index]
+            ref_px = ref_pixels[index]
+            dist = sum((px[i] - ref_px[i])**2 for i in range(3))**0.5
+            if 0 < dist <= similarity_threshold:
+                half_diff_area += 1
+            elif similarity_threshold < dist:
+                diff_area += 1
+
+        # Build an HTML image comparison
+        comparison = html_tools.build_html_tabs(
+            [
+                # TODO: Proper alt text here!
+                (
+                    "Your image:",
+                    html_tools.html_image(val, "Your image")
+                ),
+                (
+                    "Solution image:",
+                    html_tools.html_image(ref, "Solution image")
+                ),
+                (
+                    "Animation:",
+                    html_tools.html_animation(
+                        diff_anim_frames(val, ref, 10),
+                        (
+                            "An animation between your image and the"
+                            " solution image."
+                        ),
+                        delays=[500] + [100] * 10 + [500]
+                    )
+                )
+            ]
+        )
+
+        # Return a status based on the difference fraction
+        diff_fraction = diff_area / len(val_pixels)
+        half_diff_fraction = half_diff_area / len(val_pixels)
+        diff_score = diff_fraction + half_diff_fraction / 2
+
+        diff_msg = (
+            f"{diff_fraction * 100:.1f}% of pixels were"
+            f" different"
+        )
+        if half_diff_fraction > 0:
+            diff_msg += (
+                f" and {half_diff_fraction * 100:.1f}% of pixels were"
+                f" somewhat different"
+            )
+
+        if diff_score == 0:
+            return {
+                "status": "accomplished",
+                "explanation": (
+                    f"Your image and the solution image were exactly the"
+                    f" same: {comparison}"
+                )
+            }
+        elif diff_score <= allowed_differences:
+            return {
+                "status": "accomplished",
+                "explanation": (
+                    f"Your image and the solution image were almost the"
+                    f" same ({diff_msg}): {comparison}"
+                )
+            }
+        elif diff_score <= partial_allowed:
+            return {
+                "status": "partial",
+                "explanation": (
+                    f"Your image and the solution image were similar,"
+                    f" but not the same ({diff_msg}): {comparison}"
+                )
+            }
+        else:
+            return {
+                "status": "failed",
+                "explanation": (
+                    f"Your image and the solution image were not the"
+                    f" same ({diff_msg}): {comparison}"
+                )
+            }
+
+    return compare_images
+
+
 #---------------------------#
 # One-size-fits-all default #
 #---------------------------#
@@ -1786,17 +2055,29 @@ def omni_compare(val, ref):
 
     Tries its best to give a concise-ish description of differences when
     they exist.
+
+    TODO: Recursive data structures
     """
     if val == ref:
         return {
             "status": "accomplished",
             "explanation": (
-                f"Actuall value is the same as the reference"
+                f"Actual value is the same as the reference"
                 f" value.<br>\n{html_tools.dynamic_html_repr(val)}"
             )
         }
     else: # let's hunt for differences
-        if (
+        if isinstance(val, bool) and isinstance(ref, bool):
+            # Bools are isinstance of int so we need this case before the
+            # one below
+            return {
+                "status": "failed",
+                "explanation": (
+                    f"Your result ({val}) was the opposite of the"
+                    f" expected result ({ref})."
+                )
+            }
+        elif (
             isinstance(val, (int, float, complex))
         and isinstance(ref, (int, float, complex))
         ): # what if they're both numbers?
