@@ -32,11 +32,12 @@ dictionary, which is created and cached by the `Context` and supplied to
 
 import copy
 import os
-import json
 import time
 import re
 import sys
 import io
+import tempfile
+import shutil
 
 try:
     import optimism
@@ -87,7 +88,7 @@ def show_all_context_values(context):
             rep = context[key]
         else:
             try:
-                rep = json.dumps(context[key], indent=2)
+                rep = html_tools.big_repr(context[key])
             except TypeError:
                 rep = repr(context[key])
         result += f"""
@@ -147,7 +148,7 @@ class Context:
 
     - "filename": The file name (without path) of the file we're
         evaluating.
-    - "file_path": The full path to the file we're evaluating.
+    - "file_path": The full absolute path to the file we're evaluating.
     - "source": A string containing the source code of the submitted
         module.
     - "parse_errors": A list of Exception objects which were
@@ -164,10 +165,10 @@ class Context:
     - "top_scope": The top-level AST node for the submitted code.
     - "module": a loaded (submitted or solution) module
         (for a suite of tests that relies on the code).
-    - "output": a string containing output from a function call.
-        This may be the full output, or may be filtered by additional
-        Context objects to contain just part of the full output.
     - "value": arbitrary Python value returned from a function call.
+    - "output": a string containing output from tested code. This may be
+        the full output, or may be filtered by additional Context objects
+        to contain just part of the full output.
     - "trace": a trace list recording certain function calls made
         during a test, along with arbitrary state snapshots (see
         `potluck.harness.tracing_function_calls`).
@@ -178,6 +179,10 @@ class Context:
         "data" slot holds a bytes object containing the raw data in that
         format. The label slot holds a string used to label the audio in
         output to users.
+    - "output_filename": the filename for an output file written to by
+        testing code.
+    - "output_file_contents": a string containing the contents of a file
+        that was written to by tested code.
     - "expectations": a list of dictionaries defining expectations
         established using the `optimism` module (only present if that
         module is available) Each has the following keys:
@@ -222,7 +227,8 @@ class Context:
         failure_explanation=None,
         cache_values=True,
         base=None,
-        hidden=False
+        hidden=False,
+        generate_warnings=False
     ):
         """
         You must supply a description pair (topic + details), triple
@@ -261,6 +267,11 @@ class Context:
         If hidden is given as True, this Context will not show up in the
         Contexts list, and will only be visible as a context associated
         with a goal when that goal is evaluated (use with care).
+
+        If generate_warnings is set to True (default is False) then
+        issues with context creation of this context (but not with
+        creation of dependencies) will be logged as warnings in
+        `list_and_render_contexts` output.
         """
         self.description = description
         if builder is None:
@@ -274,6 +285,7 @@ class Context:
         self.cache_values = cache_values
         self.base = base or {}
         self.hidden = hidden
+        self.generate_warnings = generate_warnings
 
         self.cached_value = None
         self.cache_timestamp = None
@@ -552,7 +564,8 @@ class Context:
     def warnings(self, base_context):
         """
         Returns a list of HTML strings representing warnings generated by
-        this context. A base context is required because we need to
+        this context (excluding warnings generated by contexts this one
+        depends on). A base context is required because we need to
         generate the context value to see if there are warnings, although
         a cached value will be used in most cases. Returns an empty list
         when no warnings have been generated.
@@ -561,8 +574,15 @@ class Context:
             ctx = self.create(base_context)
             return ctx.get("warnings", [])
         except context_utils.ContextCreationError as e:
-            tb = html_tools.html_traceback(e)
-            return [ f"Error during context creation:<br>{tb}" ]
+            if isinstance(e.cause, context_utils.ContextCreationError):
+                # Error comes from a dependency, not directly from this
+                # context, so report nothing to avoid duplicating
+                # warnings
+                return []
+            else:
+                # Error comes from this context, so report it
+                tb = html_tools.html_traceback(e)
+                return [ f"Error during context creation:<br>{tb}" ]
 
 
 def add_context_numbering(all_context_objs):
@@ -791,7 +811,8 @@ def list_and_render_contexts(cgraph, base_context=None):
                 ),
                 "warnings": (
                     ctx.warnings(base_context)
-                    if base_context is not None else []
+                    if base_context is not None and ctx.generate_warnings
+                    else []
                 )
             }
         )
@@ -1060,9 +1081,9 @@ class FileContext(AutoContext):
 
             return {
                 "filename": target_filename,
-                "file_path": full_target,
+                "file_path": os.path.abspath(full_target),
                 "ref_filename": target_filename,
-                "ref_file_path": ref_target
+                "ref_file_path": os.path.abspath(ref_target)
             }
 
         # Now we can call the super-class init with the context builder
@@ -1091,13 +1112,105 @@ class FileContext(AutoContext):
 
 
 # Register a factory for a default FileContext as an on-demand option
-# for "filename" and "file_path" slots:
+# for "filename", "file_path", and the associated ref_ slots:
 AutoContext.on_demand(
     (lambda: FileContext()),
     "filename", "file_path", "ref_filename", "ref_file_path"
 )
 
 
+class SandboxContext(AutoContext):
+    """
+    Establishes a sandbox directory to be used for running all code being
+    tested, including the initial loading of the module itself.
+    """
+    def __init__(self, depends=None, hidden=True):
+        """
+        Creates a context which establishes a new unique sandbox
+        directory. The context creation function places the full path to
+        that directory in the "sandbox" context slot. A list of
+        dependencies may be provided, and hidden can be set to False if
+        desired.
+        """
+        self.dir = None
+        self.ref_dir = None
+        # TODO: Clean up these temporary directories rather than letting
+        # Python do that on shutdown...
+
+        def establish_context(prev_context):
+            """
+            Creates a new temporary directory and puts its absolute path
+            in the "sandbox" context slot. Creates a second temporary
+            directory and puts its path in the "ref_sandbox" slot.
+            """
+            tinfo = context_utils.extract(prev_context, "task_info")
+            spec = tinfo["specification"]
+
+            self.dir = tempfile.TemporaryDirectory(
+                suffix="__tmp",
+                dir=load.SANDBOX_DIR
+            )
+            self.ref_dir = tempfile.TemporaryDirectory(
+                suffix="__ref_tmp",
+                dir=load.SANDBOX_DIR
+            )
+
+            # Set up the sandboxes
+            for sb in [self.dir, self.ref_dir]:
+                # Copy helper files into the sandbox if there are any
+                # Note that we don't use symlinks here, because we don't
+                # want destructive student code to modify files outside
+                # the sandbox...
+                # TODO: Use symlinks in places where we feel safe,
+                # especially for large starter files!!!
+                helpers = context_utils.sandbox_filemap(spec)
+                if helpers is not None:
+                    for filepath in helpers:
+                        to = os.path.join(sb.name, helpers[filepath])
+                        if os.path.isdir(filepath):
+                            shutil.copytree(filepath, to)
+                        else:
+                            shutil.copy(filepath, to)
+
+            return {
+                "sandbox": os.path.abspath(self.dir.name),
+                "ref_sandbox": os.path.abspath(self.ref_dir.name)
+            }
+
+        # Now we can call the super-class init with the context builder
+        # function we just created.
+        super().__init__(
+            description=(
+                "Sandbox directories",
+                (
+                    "We will create sandbox directories for running"
+                    " your submitted code and the solution code."
+                )
+            ),
+            builder=establish_context,
+            display_product=lambda context: (
+                "Running in a sandbox"
+            ),
+            depends=depends,
+            hidden=hidden
+        )
+
+        # Finally, we can register ourselves as an auto-context for the
+        # "filename" and "file_path" slots.
+        self.register(
+            "sandbox",
+            "ref_sandbox"
+        )
+
+
+# Register a factory for a default SandboxContext as an on-demand option
+# for the "sandbox" and "ref_sandbox" slots.
+AutoContext.on_demand(
+    (lambda: SandboxContext()),
+    "sandbox", "ref_sandbox"
+)
+
+
 class CodeContext(AutoContext):
     """
     Requires "filename" and "file_path" slots (see `FileContext`), and
@@ -1111,8 +1224,6 @@ class CodeContext(AutoContext):
     will attempt to steamroll some kinds of parsing errors while
     generating associated warnings.
     """
-    # TODO: Create a default no-parse-errors Goal which uses the
-    # parse_errors context key...
     def __init__(self, depends=None, hidden=False, prep=None):
         """
         Dependencies are optional; if not specified `auto` will be used
@@ -1127,15 +1238,14 @@ class CodeContext(AutoContext):
             Establishes the following context slots based on the
             "file_path" slot, by reading the indicated file:
 
-                original_source: The raw file contents.
-                source: Possibly-edited (to steamroll syntax errors or by
-                    a prep function) file contents.
-                scope: An AST module node resulting from parsing the
-                    modified file contents.
-                top_scope: Same as above (but not designed to be
-                    modified).
-                parse_errors: A list of Exception objects that were
-                    'successfully' steamrolled by editing the source code.
+            - original_source: The raw file contents.
+            - source: Possibly-edited (to steamroll syntax errors or by a
+                prep function) file contents.
+            - scope: An AST module node resulting from parsing the
+                modified file contents.
+            - top_scope: Same as above (but not designed to be modified).
+            - parse_errors: A list of Exception objects that were
+                'successfully' steamrolled by editing the source code.
             """
             filename = context_utils.extract(prev_context, "filename")
             target = context_utils.extract(prev_context, "file_path")
@@ -1209,7 +1319,9 @@ class CodeContext(AutoContext):
                 f"The code for '{context['filename']}' (shown elsewhere)."
             ),
             depends=depends,
-            hidden=hidden
+            hidden=hidden,
+            # Errors at this level need to be reported!
+            generate_warnings=True
         )
 
         # Finally, register ourselves as an auto provider for the slots
@@ -1364,6 +1476,7 @@ class ModuleContext(AutoContext):
     """
     _src = "file_path"
     _from = "top_scope"
+    _sandbox = "sandbox"
     _to = "module"
     _to_expectations = "expectations"
     _description = (
@@ -1443,8 +1556,6 @@ class ModuleContext(AutoContext):
             """
             # Fetch the AST node that we'd like to turn into a module
             node = context_utils.extract(prev_context, self._from)
-            tinfo = context_utils.extract(prev_context, "task_info")
-            spec = tinfo["specification"]
 
             # Prefix the file name so that submitted and solution
             # modules with the same name don't collide
@@ -1492,7 +1603,10 @@ class ModuleContext(AutoContext):
                 module = load.create_module_in_sandbox(
                     node,
                     full_name,
-                    sandbox_links=load.link_mapping(spec),
+                    sandbox_dir=context_utils.extract(
+                        prev_context,
+                        self._sandbox
+                    ),
                     on_disk=src_path
                 )
                 if OPTIMISTIC:
@@ -1527,7 +1641,7 @@ class ModuleContext(AutoContext):
 
         # Figure out if we need to use automatic dependencies:
         if depends is None:
-            depends = auto(self._from)
+            depends = auto(self._from, self._sandbox)
             # Our dependency on the "filename" slot is implicit
 
         # Now we can call the super constructor
@@ -1571,6 +1685,7 @@ class SolnModuleContext(ModuleContext):
     """
     _src = "ref_file_path"
     _from = "ref_top_scope"
+    _sandbox = "ref_sandbox"
     _to = "ref_module"
     _to_expectations = "ref_expectations"
     _description = (
@@ -1789,7 +1904,7 @@ def build_context_value_displayer(
     single specific context key, and by default, will include multiple
     tabs that show the value, the reference value, and a diff of the two
     values. String values are shown as-is; non-string values are
-    converted to strings using json.dumps.
+    converted to strings using html_tools.big_repr.
 
     Set compare_ref to False to simply show the value for the specified
     key, and set include_diff to False when compare_ref is True to omit
@@ -1819,7 +1934,7 @@ def build_context_value_displayer(
                     rep = context[key]
                 else:
                     try:
-                        rep = json.dumps(context[key], indent=2)
+                        rep = html_tools.big_repr(context[key])
                     except TypeError:
                         rep = repr(context[key])
                 return f"<pre class='context-value'>{rep}</pre>"
@@ -1835,7 +1950,7 @@ def build_context_value_displayer(
                     rep = context[key]
                 else:
                     try:
-                        rep = json.dumps(context[key], indent=2)
+                        rep = html_tools.big_repr(context[key])
                     except TypeError:
                         rep = repr(context[key])
                 rep_html = f"<pre class='context-value'>{rep}</pre>"
@@ -1851,7 +1966,7 @@ def build_context_value_displayer(
                     ref_rep = context["ref_" + key]
                 else:
                     try:
-                        ref_rep = json.dumps(context["ref_" + key], indent=2)
+                        ref_rep = html_tools.big_repr(context["ref_" + key])
                     except TypeError:
                         ref_rep = repr(context["ref_" + key])
                 ref_rep_html = f"<pre class='context-value'>{ref_rep}</pre>"
@@ -1900,7 +2015,7 @@ def build_simple_context_value_displayer(
     don't need a pre wrapper and which can be displayed side-by-side
     (e.g., numbers). No diff is included, as it's presumed that any
     differences will be obvious, and values are converted to strings
-    using str() instead of json.dumps.
+    using str() instead of html_tools.big_repr.
 
     Set compare_ref to False to include only the main value.
 
