@@ -10,6 +10,8 @@ import traceback
 import difflib
 import os
 import re
+import io
+import base64
 
 from . import phrasing
 from . import file_utils
@@ -62,6 +64,11 @@ def len_as_text(html_string):
     return len(without_entities)
 
 
+#---------------------------#
+# Text wrapping/indentation #
+#---------------------------#
+
+
 def wrapped_fragments(line, width=80, indent=''):
     """
     Wraps a single line of text into one or more lines, yielding a
@@ -141,6 +148,36 @@ def wrap_text_with_indentation(text, width=80):
     return '\n'.join(wrapped)
 
 
+def indent(string, indent):
+    """
+    Indents a string using spaces. Newlines will be '\\n' afterwards.
+    """
+    lines = string.splitlines()
+    return '\n'.join(' ' * indent + line for line in lines)
+
+
+def truncate(text, limit=50000, tag='\n...truncated...'):
+    """
+    Truncates the given text so that it does not exceed the given limit
+    (in terms of characters; default 50K)
+
+    If the text actually is truncated, the string '\\n...truncated...'
+    will be added to the end of the result to indicate this, but an
+    alternate tag may be specified; the tag characters are not counted
+    against the limit, so it's possible for this function to actually
+    make a string longer if it starts out longer than the limit by less
+    than the length of the tag.
+    """
+    if len(text) > limit:
+        return text[:limit] + tag
+    else:
+        return text
+
+
+#------------------#
+# Common HTML bits #
+#------------------#
+
 def create_help_button(help_content):
     """
     Returns an HTML string for a button that can be clicked to display
@@ -376,6 +413,95 @@ def html_diff_table(
         return build_html_details(joint_title, result, classes="diff_table")
 
 
+#---------------------#
+# Multimedia elements #
+#---------------------#
+
+def html_image(image, alt_text, classes=None):
+    """
+    Given a PIL image and associated alt text, builds an HTML image tag
+    that uses a data URL to display the provided image. A list of CSS
+    class strings may be provided to include in the tag.
+    """
+    src_bytes = io.BytesIO()
+    image.save(src_bytes, format="PNG")
+    data = base64.standard_b64encode(src_bytes.getvalue()).decode("utf-8")
+    if classes is not None:
+        class_attr = 'class="' + ' '.join(classes) + '" '
+    else:
+        class_attr = ''
+    return (
+        '<img {}src="data:image/png;base64,{}" alt="{}">'
+    ).format(class_attr, data, alt_text)
+
+
+def html_animation(frames, alt_text, classes=None, delays=10, loop=True):
+    """
+    Given a list of PIL images and associated alt text, builds an HTML
+    image tag that uses a data URL to display the provided images as
+    frames of an animation. A list of CSS class strings may be provided
+    to include in the tag.
+
+    If delays is provided, it specifies the delay in milliseconds between
+    each frame of the animation. A list of numbers with length equal to
+    the number of frames may be provided to set separate delays for each
+    frame, or a single number will set the same delay for each frame.
+
+    If loop is set to false, the animation will play only once (not
+    recommended).
+    """
+    src_bytes = io.BytesIO()
+    frames[0].save(
+        src_bytes,
+        format="GIF",
+        append_images=frames[1:],
+        save_all=True,
+        duration=delays,
+        loop=0 if loop else None # TODO: Does this work?
+    )
+    data = base64.standard_b64encode(src_bytes.getvalue()).decode("utf-8")
+    if classes is not None:
+        class_attr = 'class="' + ' '.join(classes) + '" '
+    else:
+        class_attr = ''
+    return (
+        '<img {}src="data:image/gif;base64,{}" alt="{}">'
+    ).format(class_attr, data, alt_text)
+
+
+def html_audio(raw_data, mimetype, label=None):
+    """
+    Returns an HTML audio tag that will play the provided audio data,
+    using a data URL. For reasons, the data URL is included twice, so the
+    data's size will be more than doubled in the HTML output (TODO: Don't
+    do that?!). If a string is specified as the label, an aria-label
+    value will be attached to the audio element, although this will
+    usually only be accessible to those using screen readers.
+    """
+    # Construct label attribute
+    label_attr = ''
+    if label is not None:
+        label_attr = ' aria-label="{}"'.format(label)
+
+    # Encode data and construct data URL
+    data = base64.standard_b64encode(raw_data).decode("utf-8")
+    data_url = 'data:{};base64,{}'.format(mimetype, data)
+
+    # Construct and return tag
+    return '''\
+<audio controls{label}><source type="{mime}" src="{data_url}">
+(Your browser does not support the audio element. Use this link to download the file and play it using a media player: <a href="{data_url}">download audio</a>)
+</audio>'''.format( # noqa E501
+        label=label_attr,
+        mime=mimetype,
+        data_url=data_url
+    )
+
+
+#-------------------#
+# Status indicators #
+#-------------------#
+
 def status_css_class(status):
     """
     Returns the CSS class used to indicate the given status.
@@ -396,12 +522,9 @@ def build_status_indicator(status):
     )
 
 
-def indent(string, indent):
-    """
-    Indents a string using spaces. Newlines will be '\\n' afterwards.
-    """
-    lines = string.splitlines()
-    return '\n'.join(' ' * indent + line for line in lines)
+#------------------------------#
+# General Object Reprs in HTML #
+#------------------------------#
 
 
 def big_repr(thing):
@@ -429,24 +552,6 @@ def big_repr(thing):
     return base
 
 
-def truncate(text, limit=50000, tag='\n...truncated...'):
-    """
-    Truncates the given text so that it does not exceed the given limit
-    (in terms of characters; default 50K)
-
-    If the text actually is truncated, the string '\\n...truncated...'
-    will be added to the end of the result to indicate this, but an
-    alternate tag may be specified; the tag characters are not counted
-    against the limit, so it's possible for this function to actually
-    make a string longer if it starts out longer than the limit by less
-    than the length of the tag.
-    """
-    if len(text) > limit:
-        return text[:limit] + tag
-    else:
-        return text
-
-
 def build_display_box(text):
     """
     Creates a disabled textarea element holding the given text. This
@@ -478,6 +583,10 @@ def dynamic_html_repr(thing, reasonable=200, limit=10000):
         return build_display_box(truncate(rep, limit))
 
 
+#-------------------------------#
+# Formatting tracebacks as HTML #
+#-------------------------------#
+
 def html_traceback(exc=None, title=None, linkable=None):
     """
     In an exception handler, returns an HTML string that includes the
@@ -587,6 +696,10 @@ def rewrite_traceback_filenames(raw_traceback, prefix_map=None):
     return result
 
 
+#-----------------------------#
+# Function def/call templates #
+#-----------------------------#
+
 def function_def_code_tags(fn_name, params_pattern, announce=None):
     """
     Returns a tuple containing two strings of HTML code used to represent
